/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2008,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
/*
 * cpex.c
 */
 
/*
 * These routines provide a simple set of functions that allow a
 * user to copy memory from an SPU local store out to memory or
 * vice-versa..
 * Typical DMA limitations make it more difficult for the end user,
 * so for those who do not care about performance, but would rather
 * have simplicity, we provide these functions.
 *
 * DMA transfers are typically aligned on 16 byte boundries, including
 * the source address, the destination address and the number of bytes
 * to be transferred. If a copy is determined to be aligned, we will
 * use the built in mechanisms to achieve maximum performance. If a
 * copy is determined to fail an address alignment check, then we 
 * drop to a generic algorithm. DMA transfers that are note
 * aligned on 16 byte boundaries, but where the source and target
 * addresses are still aligned will utilize the efficient transfer,
 * function, but will likely suffer a performance penalty in hardware.
 * Transfers with odd sizes but proper alignment will use the efficient
 * functions as well.
 *
 * All transfers are synchronous.
 */

#include <stddef.h>
#include <stdint.h>
#include <string.h>
#include <cpex.h>

typedef union {
    unsigned long long ull;
    unsigned int ui[2];
  } ea64_t;

//#define LOCAL_DEBUGGING
#ifdef LOCAL_DEBUGGING
#include <stdio.h>
#define PRINTF(...) { printf(__VA_ARGS__); fflush(stdout); }
#else
#define PRINTF(...) {}
#endif

typedef unsigned int 	tagid_t;

#if defined(__xlC__) && (__xlC__ < 0x0a01)
/* versions of xlc/xlC before 10.1 do not have this */
/*  defined in their spu_mfcio.h file, so we'll do it */
#ifdef SPU_MFCIO_INTERRUPT_SAFE
static inline uint32_t 
mfc_begin_critical_section(void)
{
    uint32_t MachStat = spu_readch(SPU_RdMachStat);
    spu_idisable();
    return MachStat;
}

static inline void
mfc_end_critical_section(uint32_t mach_stat)
{
    if (mach_stat & 0x1)
	spu_ienable();
}

#else /* NOT interrupt safe */

static inline uint32_t 
mfc_begin_critical_section(void)
{
    /* nothing - not interrupt safe */
    return 0;
}

static inline void
mfc_end_critical_section(uint32_t mach_stat __attribute__ ((unused)))
{
    /* nothing - not interrupt safe */
}

#endif /* need interrupt safe */
#endif /* correct xlc version */

static inline void
__copy_mfcdma(uint32_t ls, uint64_t ea, uint32_t len,
                   uint32_t op, uint32_t tagid) 
{
    uint32_t mach_stat = mfc_begin_critical_section();
    ea64_t ms = {.ull = ea};
    spu_mfcdma64((char *)ls, ms.ui[0], ms.ui[1], len, tagid, op);
    mfc_end_critical_section(mach_stat);
}


/* __copy_wait 
 *    internal routine to serialize with 
 *    pending asynchronous copy operations.
 */
static inline void __copy_wait(uint32_t mask)
{
  uint32_t old_tag_mask;

  /* read and save the current tag mask so we don't corrupt
   * the state of the tag masks any code calling us
   */
  old_tag_mask = spu_readch(MFC_RdTagMask);

  /* set our mask */
  spu_writech(MFC_WrTagMask, mask);
  spu_writech(MFC_WrTagUpdate, MFC_TAG_UPDATE_ALL);

  /* wait for the tag to return */
  spu_readch(MFC_RdTagStat);

  /* restore the previous mask */
  spu_writech(MFC_WrTagMask, old_tag_mask);
}

/**
 * copy_wait - Synchronize with pending asynchronous copy operations.
 * @tagid: tag identifier for asynchronous copy operation.
 *
 * Synchronize with pending asynchronous copy operations.
 * This operation effectively waits for completion of any
 * operations with the given @tagid.  If @tagid is greater
 * than %MAX_TAG_ID, then this routine will wait for all
 * pending asynchronous transfers to complete.
 */
void copy_wait(tagid_t tagid)
{
  uint32_t mask = (tagid > 0x1f) ? ~0 : (1 << tagid);
  
  __copy_wait(mask);
}

/*****************************************************************************/

/*
 * Function: __copy_shift
 * Args:
 *      buf - User buffer being DMA'd.  Data will be shifted within this
 *            buffer.
 *      len - Length of the user buffer to re-use in transfer
 *      off - Calculated offset into user buffer to be shifted.  The offset in
 *            should guarantee that we are not manipulating a partial 16 byte
 *            portion of the main buffer.
 *      shift - Number of bytes to shift data in user buffer
 *      fbuf - Pointer to temp buffer to be used for DMAing front portion
 *             of user data.
 *      flen - Number of bytes to be copied to front-end buffer
 *      foff - Offset into front buffer to copy data to.  This should
 *             be same as remote buffer offset.
 *      tbuf - Pointer to temp buffer to be used for DMAing tail portion
 *             of user data.
 *      tlen - Number of bytes to be copied to tail-end buffer
 *      toff - Offset into tail buffer to copy data to.  This is calculated 
 *             based on the number of bytes copied to the front buf plus the
 *             number preserved in the main buffer.
 */
static void
__copy_shift(char *buf, uint32_t len, uint32_t off, int32_t shift, char *fbuf,
             uint32_t flen, uint32_t foff, char *tbuf, uint32_t tlen,
             uint32_t toff)
{
    unsigned int i;

    /*
     * A negative shift is a shift right.  Byte copy the front and tail of the
     * main buffer based on the indicated byte counts.
     */
    if (shift < 0) {
        for (i = 0; i < flen; i++)
            fbuf[foff+i] = buf[i];

        for (i = 0; i < tlen; i++)
            tbuf[toff+i] = buf[len+flen+i];
    }

    /*
     * Otherwise, the buffer is being shifted left. 
     * Start by restoring the front of the original buffer.  This step is
     * actually not necessary when we are restoring a buffer, as we don't
     * actually disrupt the front of the main buffer.
     *
     * The tail is restored/shifted after the middle portion because the data
     * must be left in tact while we are handling the middle.
     */
    else {
        for (i = 0; i < flen; i++)
            buf[i] = fbuf[foff+i];
    }

    /*
     * On small transfers, we may have done everything up above.  If not, then
     * we have a main buffer length that we will vectorize and rotate.
     */
    if (len) {
        vector unsigned char *vbuf;
#define TENS {16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16};
#define LO_ENUM {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15}
        vector unsigned char tens = TENS;
        vector unsigned char rot = LO_ENUM;
        int v, vcnt, vshift, voff;

	/*
         * The vector ops will truncate the vector pointers, so we may as well
         * do it first so we know where we are.  This may not be necessary
         * when we are using 128 byte buffer alignment.
         * NOTE: vbuf points to the first vector of the working buffer (offset
         * by 'off'), not the first vector of the incoming vector.
	 */
        vbuf = (vector unsigned char *)(((uint32_t)&buf[off]) & ~0xF);

	/*
         * Determine how many vectors between the truncated start and the end.
	 */
        vcnt = (len+15) >> 4;

	/*
         * We could be shifting by more than a quadword in length, so we need
         * to break the shift into 2 components:
         *   voff   => # of vector indices to shift
         *             This offset is managed as a positive value that
         *             indicates how many vectors to the left or right to
         *             copy the calculated vector shuffle.
         *   vshift => # of bytes within the vector to shift.
         *             This shift value must maintain the same polarity as the
         *             incoming shift amount and is used to determine how many
         *             bytes left or right to shuffle a particular vector.
         *
         *    +-- buf
         *    V
         *  +--------+--------+--------+--------+--------+--------+
         *  |        |        |        |        |        |        |
         *  +--------+--------+--------+--------+--------+--------+
         *    |                                          ^
         *    +--------->--------- shift --------->------+
         *    |      ^                                   ^
         *    +vshift+                                   |
         *           |                                   |
         *           +-------------voff------------------+
         *
         *  For example, if the shift amount was 68 bytes, this would be
         *  broken into a 4 byte intra-vector shift and a 4 vector offset
         *  (64/16), or vshift=(+/-)4 and voff=4.
         *
         * Again, a negative shift is a shift right.
	 */
        if (shift < 0) {
	    /*
             * The vector offset is a positive vector index shift to the
             * right which is equivalent to a 16 byte vector shift.
	     */
            voff = (-shift) >> 4;
            
	    /*
             * Maintain the polarity of the original shift and truncate it
             * into a sub-quadword amount.
	     */
            vshift = shift + (voff << 4);

	    /*
             * Rotate a vector indices mask right (inverted left, 16+vshift) 
             * to create a single vector shuffle mask.  The result is a mask
             * that has the correctly ordered byte pattern from the first
             * shuffle vector (left-most bytes).
             * 
             * Then, OR in a mask of 0x10s shifted the same amount as above to 
             * the right.  This will convert the right-most shuffle bytes to
             * the second shuffle vector.
             * 
             * For example, in a 3 byte shift to the right...
             *  index mask  =>  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
             *  rotate R 3  =>  D  E  F  0  1  2  3  4  5  6  7  8  9  A  B  C
             *  0x10s R 3   =>  0  0  0 10 10 10 10 10 10 10 10 10 10 10 10 10
             *  OR          =>  D  E  F 10 11 12 13 14 15 16 17 18 19 1A 1B 1C
             */
            rot = spu_rlqwbyte(rot, 16+vshift);
            rot = spu_or(rot, spu_rlmaskqwbyte(tens, vshift));

            /*
	     * Shifting right starts on the right-most vector and works
             * toward the left-most vector.  It must be done in this manner to 
             * preserve the left-most data, which is used in calculating the 
             * data to the right.  The voff value is used to decide which 2 
             * earlier vectors are used to calculate the shuffle pattern on the
             * resultant vector.
             * NOTE: The vector pointers supplied to the shuffle routine could
             * be behind vbuf as earlier data is needed.
	     */
            for (v = vcnt-1; v >= 0; v--)
                vbuf[v] = spu_shuffle(vbuf[v-voff-1], vbuf[v-voff], rot);
        } else {
            /*
	     * The vector offset is a positive vector index shift to the left
             * which is equivalent to a 16 byte vector shift.
	     */
            voff = (shift) >> 4;

            /*
	     * The intra-vector shift will be a positive, sub-quadword amount
             * of bytes.
	     */
            vshift = shift - (voff << 4);

            /*
	     * Similar to the above explanation, we will be rotating a vector
             * indices mask to the left by the intra-vector shift amount
             * (vshift).  This will give us the left-most bytes of the shuffle
             * pattern, indicating which bytes to use in the first shuffle
             * vector.  
             *
             * Then, we OR in a mask of 0x10s shifted an inverted amount to the
             * right to cover the previously shifted mask.  This will cover
             * the right-most shuffle bytes to the second shuffle vector.
             *
             * For example, in a 3 byte shift to the left...
             *  index mask  =>  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
             *  rotate L 3  =>  3  4  5  6  7  8  9  A  B  C  D  E  F  0  1  2
             *  0x10s R 3   =>  0  0  0  0  0  0  0  0  0  0  0  0  0 10 10 10
             *  OR          =>  3  4  5  6  7  8  9  A  B  C  D  E  F 10 11 12
             */
            rot = spu_rlqwbyte(rot, vshift);
            rot = spu_or(rot, spu_rlmaskqwbyte(tens, vshift-16));

            /*
	     * Shifting left starts on the left-most vector and works toward
             * the right-most vector.  It must be done in this manner to
             * preserve the right-most data, which is used in calculating the
             * data to the left.  The voff value is used for determining which
             * vector to the left the shuffle results are written to. 
             * NOTE: The resultant vector index can be negative.  This is
             * because vbuf points to the first vector at 'off' offset into
             * buf.  
	     */
            for (v = 0; v < vcnt; v++)
                vbuf[v-voff] = spu_shuffle(vbuf[v], vbuf[v+1], rot);
        } /* if / else shift < 0 */
    } /* if len */

    /*
     * The data at the end of the main buffer has been shifted, so it is now
     * safe to restore the tail data this is a left shift.
     */
    if (shift > 0) {
        for (i = 0; i < tlen; i++)
            buf[len+flen+i] = tbuf[i+toff];
    }
} /* __copy_shift */

void __copy_to_from_aligned(uint32_t ls, uint64_t ea, size_t len, 
					uint32_t op, tagid_t tagid)
{
    unsigned int mask[] = {1, 2 ,4 ,8, ~0xF, ~0x7, ~0x3, ~0x1, -1};
    unsigned int sz;
    unsigned int i;

    /*
     * Start by DMAing any sub-16 byte pieces off the front.  Our goal upon
     * completion of this loop is to have a 16 byte aligned address.  This is
     * done by aligning the address by doing a naturally aligned DMA based on
     * the address.  We loop through the front half of the above mask, doing
     * DMAs (in order) of 1 byte, 2 bytes, 4 bytes, and 8 bytes.
     */
    for (i = 0; i < 4; i++) {
        sz = (uint32_t)ls & mask[i];
        if (sz <= len) {
            __copy_mfcdma(ls, ea, sz, op, tagid);
            ea += sz;
            ls += sz;
            len -= sz;
            if (!len) return;
        }
    }

    /*
     * At this point we possibly did a series of smaller DMAs in order to have
     * a 16-byte address to work from.  Now, we can do larger DMAs, followed
     * by smaller ones.  Unlike the above loop, this loop is based on the
     * remaining length, rather than on address alignment.
     * Cell has a maximum DMA size of 16k, so we start by peeling off any 16k
     * chunks.  Then we do the largest multiple of 16 bytes, followed by
     * picking up the scraps: 8 bytes, 4 bytes, 2 bytes, 1 byte.
     */
    do {
        sz = ((len & mask[i]) >= MFC_MAX_DMA_SIZE) ? MFC_MAX_DMA_SIZE : (len & mask[i]);
        if (sz) {
            __copy_mfcdma(ls, ea, sz, op, tagid);
            ea += sz;
            ls += sz;
            len -= sz;
        }
        i += !(len & mask[i]);
    } while (len > 0);

} /* __copy_to_from_aligned */

static inline void __copy_to_from_unaligned(uint32_t ls, uint64_t ea, size_t totlen, 
					uint32_t op, tagid_t tagid)
{

    /*
     * Determine the remote and local alignments based on the above alignment
     * choice.
     */
    uint32_t Ar, Al;

    uint32_t len, off, flen, foff, tlen, toff;
    uint32_t shift;
    vector unsigned char tempdata[48];
    char *fbuf = (char *)(((uint32_t)tempdata + (CPEX_ALIGN-1)) & ~(CPEX_ALIGN-1));
    char *tbuf = (char *)((uint32_t)fbuf + (CPEX_ALIGN*2));

    /*
     * Determine the remote and local alignments based on the above alignment
     * choice.
     */
    Ar = ea & (CPEX_ALIGN-1);
    Al = (uint32_t)ls & (CPEX_ALIGN-1);

    /*
     * The shift amount is the number of bytes we must shift right the local
     * buffer to match the remote side alignment.  
     */
    shift = (Al < Ar) ? Ar - Al : (CPEX_ALIGN-Al)+Ar;
    shift &= (CPEX_ALIGN-1);

    /*
     * Start with a front-end buffer size equal to the number of
     * non-quadword aligned bytes on the front of the local buf.
     * The front end buffer manages less than a quadword to minimize RMW ops.
     * The remaining quadword data is managed through the data shuffle.
     */
    flen = 16-((uint32_t)ls & 0xF);

    /*
     * The offset of the front is always the same as the remote buffer
     * alignement.
     */
    foff = Ar;

    /*
     * If we are doing a PUT, then we handle the front differently.  Since
     * we are shifting our data right, we may end up shifting all the
     * "unaligned" bytes into the first full QW.  Full QW data can be
     * rotated and the shifted amount will be caught in the tail end
     * buffer.  For this reason, we can decrement the front end size by
     * shift or set it to 0 if the whole amount is shifted.
     */
    if (op & MFC_PUT_CMD)
        flen = (flen > shift) ? flen - shift : 0;

    /*
     * It is possible that the whole DMA length falls in the unaligned
     * section.  If so, truncate our buffer.
     */
    flen = (flen > totlen) ? totlen : flen;

    /*
     * The starting offset of the main buffer that is being reused depends
     * on where we left off with the front portion.  If no front side
     * buffer is used (flen = 0), then we need to be at least aligned with
     * the remote side, which is ls[off].  If we did some front side
     * transfer, then we need to add in the front length.
     * This is used for offsetting from the beggining of the local side
     * buffer, so it is not truncated.
     */
    off = flen + shift;

    /*
     * We pick up the scraps on the end.  This is the partial alignment bytes 
     * on the end of the local buf plus the bytes shifted out of the main
     * buffer.
     */
    tlen = ((Al + totlen) & (CPEX_ALIGN -1)) + shift;

    /*
     * Make sure that the calculated tail length is not more that the
     * remaining bytes.
     */
    tlen = (tlen > (totlen - flen)) ? (totlen-flen) : tlen;

    /*
     * Anything we did not DMA in the front or tail will be done within
     * the main buffer.  This is the amount we can rotate.
     */
    len = totlen - flen - tlen;

    /*
     * The tail offset depends on the number of bytes DMA'd in the front
     * and main buffers.  The tail offset is the number of bytes past the
     * remote buffer alignement.  We truncate the alignment, since we are
     * using a temporary buffer.
     */
    toff = (Ar + flen + len) & (CPEX_ALIGN -1);

    /*
     * If this is an outgoing buffer, we need to shift the data into place
     * prior to sending it out.  This is a shift right, which is represented
     * as a negative shift.
     *
     * Note: This makes PUTs more expensive than GETs on the SPE.
     */
    if (op & MFC_PUT_CMD)
        __copy_shift((char *)ls, len, off, -shift, fbuf, flen, foff, tbuf, tlen, toff);

    /*
     * Initiate the main buffer DMA.
     *
     * At this point we know the necessary offsets to perform properly aligned
     * DMAs.  This may have required some calculations and adjustments
     * performed above.
     *
     * In the case of a PUT, we are sending a buffer with data shifted to
     * match the remote side buffer.
     * In the case of a GET, we are receiving into a location that matches the
     * remote side alignment.
     * In both cases, we are only DMAing the precalculated amount, which could
     * be zero if the alignment and size don't work.
     */
    __copy_to_from_aligned((uint32_t)&((char *)ls)[off], ea+flen, len, op, tagid);

    /*
     * Pick-up the pieces
     *
     * If we had to shift things then we possibly have a front end buffer
     * and/or a tail end buffer.  We will DMA them now.  Again, we calculated
     * the appropriate offsets above, which guarantees that the source and
     * destination buffers are aligned.
     */
    __copy_to_from_aligned((uint32_t)&fbuf[foff], ea, flen, op, tagid);
    __copy_to_from_aligned((uint32_t)&tbuf[toff], ea + flen + len, tlen, op, tagid);

    /*
     * Since we had to mess with the users main buffer data, then it is in
     * an inconsistent state.  Prior to returning we have to make sure we
     * restore the users buffer, but before we can do that, we have to
     * insure that the DMA is complete.  So we must wait.
     */
    __copy_wait(1 << tagid);

    /*
     * Move the local data back into its correct location.  On a put this
     * is simply us restoring the original data state.  On a get, we are
     * moving the received data into it's expected position.
     */
    __copy_shift((char *)ls, len, off, shift, fbuf, flen, foff, tbuf, tlen, toff);

} /* __copy_to_from_unaligned */


/*****************************************************************************/
/**
 * These are the external interfaces
 */
/*****************************************************************************/

/**
 * copy_to_ls - Copy from main memory to LS - dma get
 * @to: pointer to local storage area buffer (destination).
 * @from: handle to effective address buffer (source).
 * @n: number of bytes to be copied.
 *
 * This operation is synchronous.
 */
size_t copy_to_ls(uint32_t to, uint64_t from, size_t n)
{
  /* reserve a tag for our usage */
  const tagid_t tagid = mfc_tag_reserve();

  uint32_t aligned = ARE_ALIGNED(from, to);
  PRINTF("cpex: copy_to_ls - aligned returns 0x%x, ea=0x%x, ls=0x%x\n", 
  	aligned, ((uint32_t) ((from) & (CPEX_ALIGN-1))), ((uint32_t) (to) & (CPEX_ALIGN-1)));
  
  /* do the copy */
  if (aligned)
    __copy_to_from_aligned(to, from, n, MFC_GET_CMD, tagid);
  else
    __copy_to_from_unaligned(to, from, n, MFC_GET_CMD, tagid);
  
  /* extra sync only required for successful aligned transfers */
  if (aligned) {
    __copy_wait(1 << tagid);
  }
  
  /* release the tag that we had reserved */
  (void)mfc_tag_release(tagid);
  
  return n;
}

/**
 * copy_to_ls_aligned - Copy from main memory to LS - dma get
 * Same as copy_to_ls(), but much smaller and more efficient. 
 * Assumes caller has checked alignment.
 * @to: handle to effective address buffer (destination).
 * @from: pointer to local storage area buffer (source).
 * @n: number of bytes to be copied.
 *
 * This operation is synchronous.
 */
size_t copy_to_ls_aligned(uint32_t to, uint64_t from, size_t n)
{
  /* reserve a tag for our usage */
  const tagid_t tagid = mfc_tag_reserve();

  /* do the copy */
  __copy_to_from_aligned(to, from, n, MFC_GET_CMD, tagid);

  /* wait for transfer to complete */
  __copy_wait(tagid);

  /* release the tag that we had reserved */
  (void)mfc_tag_release(tagid);

  return n;
}

/**
 * copy_from_ls - Copy to main memory from LS - dma put
 * @to: handle to effective address buffer (destination).
 * @from: pointer to local storage area buffer (source).
 * @n: number of bytes to be copied.
 *
 * This operation is synchronous.
 */
size_t copy_from_ls(uint64_t to, uint32_t from, size_t n)
{
  /* reserve a tag for our usage */
  const tagid_t tagid = mfc_tag_reserve();

  uint32_t aligned = ARE_ALIGNED(to, from);
  PRINTF("cpex: copy_from_ls - aligned returns 0x%x, ea=0x%x, ls=0x%x\n", 
  	aligned, ((uint32_t) ((to) & (CPEX_ALIGN-1))), ((uint32_t) (from) & (CPEX_ALIGN-1)));

  /* do the copy */
  if (aligned)
    __copy_to_from_aligned(from, to, n, MFC_PUT_CMD, tagid);
  else
    __copy_to_from_unaligned(from, to, n, MFC_PUT_CMD, tagid);
  
  /* extra sync only required for successful aligned transfers */
  if (aligned) {
    __copy_wait(1 << tagid);
  }

  /* release the tag that we had reserved */
  (void)mfc_tag_release(tagid);

  return n;
}

/**
 * copy_from_ls_aligned - Copy to main memory from LS - dma put
 * Same as copy_from_ls(), but much smaller and more efficient. 
 * Assumes caller has checked alignment.
 * @to: handle to effective address buffer (destination).
 * @from: pointer to local storage area buffer (source).
 * @n: number of bytes to be copied.
 *
 * This operation is synchronous.
 */
size_t copy_from_ls_aligned(uint64_t to, uint32_t from, size_t n)
{
  /* reserve a tag for our usage */
  const tagid_t tagid = mfc_tag_reserve();

  /* do the copy */
  __copy_to_from_aligned(from, to, n, MFC_PUT_CMD, tagid);

  /* wait for transfer to complete */
  __copy_wait(tagid);

  /* release the tag that we had reserved */
  (void)mfc_tag_release(tagid);

  return n;
}

