/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <spu_mfcio.h>
#include <race_check.h>
#include "race_check_global.h"

/* race_check_dma
 * ==============
 * Verify that the specified DMA does not produce a race condition. Also, 
 * add the DMA to race check data base so that is can be race checked with
 * future DMA or memory accesses.
 *
 * The DMA is specified by the parameters:
 *  start - specifies the starting local store address to be transferred.
 *  size  - specifies the number of bytes in the transfer.
 *  cmd   - specifies the MFC DMA command of the transfer. This does not
 *          include the tid or rid.
 *  tag   - specifies the tag group ID of the DMA transfer.
 *  filename - pointer to the name of the file that initiated the tranfer.
 *          This optional parameter is only used when RACE_CHECK_VERBOSE 
 *          is defined.
 *  line  - file line number that initiated the transfer. This optional 
 *          parameter is only used when RACE_CHECK_VERBOSE is defined.
 */

void race_check_dma(volatile void *start, 
		    unsigned int size, 
		    unsigned int cmd, 
		    unsigned int tag
		    RACE_CHECK_VERBOSE_PARMS1)
{
  int i;
  int entry;
  unsigned int flags;

#if RACE_CHECK_VERBOSE && RACE_CHECK_DEBUG
  fprintf(RACE_CHECK_STREAM, "RACE CHECK - %s (%s:%d start=%p size=0x%x cmd=0x%x tag=%d)\n", __PRETTY_FUNCTION__, filename, line, start, size, cmd, tag);
#endif

  /* Add DMA transfer to the data base 
   */
  entry = _race_check_find_db_entry();

#if RACE_CHECK_VERBOSE
  race_check_db[entry].filename = filename;
  race_check_db[entry].line = line;
#endif
  race_check_db[entry].start = (void *)start;
  race_check_db[entry].size = size;
  race_check_db[entry].cmd = cmd;
  race_check_db[entry].tag = tag;
  race_check_db[entry].barrier_tags = 0;
  race_check_db[entry].flags = FLAG_INFLIGHT;

  if (size) {
    for (i=0; i<entry; i++) {
      flags = race_check_db[i].flags;

      if (flags & FLAG_INFLIGHT) { 
	if (IS_NON_QUEUED_ATOMIC_CMD(cmd) || IS_NON_QUEUED_ATOMIC_CMD(race_check_db[i].cmd) || 
	    (((flags & FLAG_BARRIERED_NON_TAG_SPECIFIC) == 0) && 
	     ((tag != race_check_db[i].tag) || (((flags & FLAG_BARRIERED) == 0) && ((cmd & (MFC_BARRIER_ENABLE | MFC_FENCE_ENABLE)) == 0))))) {
	  /* Check for possible race condition between two DMAs. These occur
	   * when:
	   *   two transfers overlap AND
	   *   the transfers are GET-GET, GET-PUT, or PUT-GET (PUT-PUT is ok) AND
	   *   the tag IDs are not equal OR the first transfer has not been barrierd OR this transfer is not barriered or fenced.
	   */
	  /* Check DMA target */
	  if ((start < (race_check_db[i].start + race_check_db[i].size)) &&
	      ((start+size) > race_check_db[i].start) && 
	      (race_check_db[i].size) &&
	      IS_EITHER_GET_CMD(cmd, race_check_db[i].cmd)) {
#if RACE_CHECK_VERBOSE
	    fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d)\n",
		    filename, line, cmd, start, size, tag, 
		    race_check_db[i].filename, race_check_db[i].line, race_check_db[i].cmd, race_check_db[i].start, race_check_db[i].size, race_check_db[i].tag);
#else
	    fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMA_ERROR);
#endif
	  }
	  
	  /* Check list element */
	  if (flags & FLAG_DMA_LIST) {
	    if ((start < ((void *)race_check_db[i].list + race_check_db[i].list_size)) &&
		((start+size) > (void *)race_check_db[i].list) &&
		IS_GET_CMD(cmd)) {
#if RACE_CHECK_VERBOSE
	      fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a list array used in the DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x)\n",
		      filename, line, cmd, start, size, tag,
		      race_check_db[i].filename, race_check_db[i].line, race_check_db[i].cmd, race_check_db[i].start, race_check_db[i].size);
#else
	      fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMA_LIST_ERROR);
#endif
	    }
	  }
	}
      }
    }
  }
}
