/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <spu_mfcio.h>
#include <race_check.h>
#include "race_check_global.h"

/* race_check_dma_list
 * ===================
 * Verify that the specified DMA list does not produce a race condition. Also, 
 * add the DMA list to race check data base so that is can be race checked with
 * future DMA or memory accesses.
 *
 * The DMA is specified by the parameters:
 *  start - specifies the starting local store address to be transferred.
 *  list  - pointer to the DMA list element array of size bytes long.
 *  list_size  - specifies the number of bytes in the list element array. Must
 *          be an integral multiple of 8.
 *  cmd   - specifies the MFC DMA command of the transfer. This does not
 *          include the tid or rid.
 *  tag   - tag group ID of the DMA transfer.
 *  filename - pointer to the name of the file that initiated the tranfer.
 *          This optional parameter is only used when RACE_CHECK_VERBOSE 
 *          is defined.
 *  line  - file line number that initiated the transfer. This optional 
 *          parameter is only used when RACE_CHECK_VERBOSE is defined.
 */

void race_check_dma_list(volatile void *start,
			 volatile void *list,
			 unsigned int req_list_size,
			 unsigned int cmd,
			 unsigned int tag
			 RACE_CHECK_VERBOSE_PARMS1)
{
  int i;
  int overwrite=0;
  unsigned int size = 0;
  unsigned int list_size;	/* List size up through first stall */
  unsigned int element_size;
  unsigned int flags = FLAG_INFLIGHT | FLAG_DMA_LIST;
  int entry;
  mfc_list_element_t *element;
  

#if RACE_CHECK_VERBOSE && RACE_CHECK_DEBUG
  fprintf(RACE_CHECK_STREAM, "RACE CHECK - %s (%s:%d start=%p list=%p list_size=0x%x cmd=0x%x tag=%d)\n", __PRETTY_FUNCTION__, filename, line, start, list, list_size, cmd, tag);
#endif

  /* Compute the local store size of the transfer, also make sure
   * that a GETL doesn't overwrite its own list array.
   */
  overwrite = 0;

  list_size = 0;
  if (req_list_size > 0) {
    while (1) {
      element = (mfc_list_element_t *)(list + list_size);

      element_size = element->size;
      if ((element_size < 16) && (element_size != 0)) element_size = 16;

      list_size += 8;

      if (((start+size) < ((void *)list+req_list_size)) &&
	  ((void *)list+list_size < (start+size+element_size)))  {
	overwrite = 1;
      }
      size += element_size;

      /* Terminate the loop early so that we don't mark the list as stallable if the
       * flag is set on the last element 
       */
      if (list_size >= req_list_size) break;

      /* Restrict the transfer to the first stall and notify flag */
      if (element->notify) {
	flags |= FLAG_STALLABLE;
	break;
      }
    }
  }

  if (overwrite && IS_GETL_CMD(cmd)) {
#if RACE_CHECK_VERBOSE
    fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for DMA list issued in %s:%d (cmd=0x%x start=%p list=%p list_size=0x%x tag=%d ) which overwrote its own list element array\n",
	    filename, line, cmd, start, list, req_list_size, tag);
#else
    fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_LIST_OVERWRITE);
#endif
  }

  /* Add DMA list transfer to the data base 
   */
  entry = _race_check_find_db_entry();

#if RACE_CHECK_VERBOSE
  race_check_db[entry].filename = filename;
  race_check_db[entry].line = line;
#endif
  race_check_db[entry].start = (void *)start;
  race_check_db[entry].size = size;
  race_check_db[entry].cmd = cmd;
  race_check_db[entry].tag = tag;
  race_check_db[entry].list = (void *)list;
  race_check_db[entry].list_size = list_size;
  race_check_db[entry].req_list_size = req_list_size;
  race_check_db[entry].barrier_tags = 0;
  race_check_db[entry].flags = flags;

  if (size) {
    for (i=0; i<entry; i++) {
      flags = race_check_db[i].flags;
      
      if (flags & FLAG_INFLIGHT) {
	/* If this command is not barriered or fenced and this command is stallable
	 * and a previous command is stallable with the same tag, then flag a race condition
	 * in that users will be unable to detected which command stalled because they
	 * are not ordered.
	 */
	if (((cmd & (MFC_FENCE_ENABLE | MFC_BARRIER_ENABLE)) == 0) && (flags & FLAG_STALLABLE) && (tag == race_check_db[i].tag) && ((flags & FLAG_BARRIERED) == 0)) {
#if RACE_CHECK_VERBOSE
	  fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for stallable DMA list issued in %s:%d (cmd=0x%x start=%p tag=%d) with stallable DMA list issued in %s:%d (cmd=0x%x start=%p)\n",
		  filename, line, cmd, start, tag,
		  race_check_db[i].filename, race_check_db[i].line, race_check_db[i].cmd, race_check_db[i].start);
#else
	  fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_STALLABLE_ERROR);
#endif
	}
	
	if (IS_NON_QUEUED_ATOMIC_CMD(race_check_db[i].cmd) || 
	    (((flags & FLAG_BARRIERED_NON_TAG_SPECIFIC) == 0) && 
	     ((tag != race_check_db[i].tag) || (((flags & FLAG_BARRIERED) == 0) && ((cmd & (MFC_BARRIER_ENABLE | MFC_FENCE_ENABLE)) == 0))))) {
	  
	  /* Check for possible race condition between two DMAs. These occur
	   * when:
	   *   two transfers overlap AND
	   *   the transfers are GET-GET, GET-PUT, or PUT-GET (PUT-PUT is ok) AND
	   * OR
	   *   previous transfer overlaps this list element array AND
	   *   previous transfer is a GET (or GETL).
	   */
	  /* Check DMA target */
	  if (((start < (race_check_db[i].start + race_check_db[i].size)) &&
	       ((start+size) > race_check_db[i].start) &&
	       (race_check_db[i].size) &&
	       IS_EITHER_GET_CMD(cmd, race_check_db[i].cmd)) ||
	      (((void *)list < (race_check_db[i].start + race_check_db[i].size)) &&
	       (((void *)list+list_size) > race_check_db[i].start) &&
	       IS_GET_CMD(race_check_db[i].cmd))) {
#if RACE_CHECK_VERBOSE
	    fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d)\n",
		    filename, line, cmd, start, size, tag, 
		    race_check_db[i].filename, race_check_db[i].line, race_check_db[i].cmd, race_check_db[i].start, race_check_db[i].size, race_check_db[i].tag);
#else
	    fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_ERROR);
#endif
	  }
	  
	  /* Check list element */
	  if (flags & FLAG_DMA_LIST) {
	    if ((start < ((void *)race_check_db[i].list + race_check_db[i].list_size)) &&
		((start+size) > (void *)race_check_db[i].list) &&
		(race_check_db[i].list_size) &&
		IS_GET_CMD(cmd)) {
#if RACE_CHECK_VERBOSE
	      fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a list array used in the DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d)\n",
		      filename, line, cmd, start, size, tag,
		      race_check_db[i].filename, race_check_db[i].line, race_check_db[i].cmd, race_check_db[i].start, race_check_db[i].size, race_check_db[i].tag);
#else
	      fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_LIST_ERROR);
#endif
	    }
	  }
	}
      }
    }
  }
}
