/* --------------------------------------------------------------  */
/* (C)Copyright 2006,2008,                                         */
/* International Business Machines Corporation                     */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <spu_mfcio.h>
#include <race_check.h>
#include "race_check_global.h"

/* race_check_list_stall_ack
 * =========================
 * Acknowledge a DMA list that has been stalled and resume the transfer.
 *
 * tag - specifies the tag group containing the DMA list to be acknowledged and resumed.
 */

void race_check_list_stall_ack(unsigned int tag)
{
  int i, j;
  int overwrite=0;
  unsigned char cmd;
  unsigned int size = 0;
  unsigned int req_list_size;
  unsigned int list_size;	
  unsigned int element_size;
  unsigned int flags, flags_j, flags_older = 0;
  mfc_list_element_t *element;
  void *start, *list;

  
#if RACE_CHECK_VERBOSE && RACE_CHECK_DEBUG
  fprintf(RACE_CHECK_STREAM, "RACE CHECK - %s (tag=%d)\n", __PRETTY_FUNCTION__, tag);
#endif

  /* Scan the database for the first stalled DMA list with the specified tag.
   */
  for (i=0; i<RACE_CHECK_DB_SIZE; i++) {
    flags = race_check_db[i].flags;

    if ((race_check_db[i].tag == tag) && (flags & FLAG_STALLED)) {
      /* Found. Compute new DMA parameters.
       */
      start = race_check_db[i].start + race_check_db[i].size;
      list = race_check_db[i].list + race_check_db[i].list_size;
      req_list_size = race_check_db[i].req_list_size - race_check_db[i].list_size;
      cmd = race_check_db[i].cmd;
      
      list_size = 0;
      if (req_list_size > 0) {
	while (1) {
	  element = (mfc_list_element_t *)(list + list_size);

	  element_size = element->size;
	  if ((element_size < 16) && (element_size != 0)) element_size = 16;

	  list_size += 8;

	  if (((start+size) < ((void *)list+req_list_size)) &&
	      ((void *)list+list_size < (start+size+element_size)))  {
	    overwrite = 1;
	  }
	  size += element_size;

	  /* Terminate the loop early so that we don't mark the list as stallable if the
	   * flag is set on the last element 
	   */
	  if (list_size >= req_list_size) break;

	  /* Restrict the transfer to the next stall and notify flag */
	  if (element->notify) {
	    race_check_db[i].flags |= FLAG_STALLABLE;
	    break;
	  }
	}
      }
      
      /* Update the data base entry for the DMA resuming.
       */
      race_check_db[i].start = start;
      race_check_db[i].size = size;
      race_check_db[i].list = list;
      race_check_db[i].list_size = list_size;
      race_check_db[i].req_list_size = req_list_size - list_size;
      race_check_db[i].flags = 0;

      if (overwrite && IS_GETL_CMD(cmd)) {
#if RACE_CHECK_VERBOSE
	fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected an acknowledged DMA list %s:%d (cmd=0x%x start=%p list=%p list_size=0x%x tag=%d ) overwrote its own list element array\n",
		race_check_db[i].filename, race_check_db[i].line, cmd, start, list, list_size, tag);
#else
	fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_LIST_OVERWRITE);
#endif
      }

      /* Search the database for a race condition between this DMA being resumed and the others
       * that are already in-flight.
       */
      if (size) {
	for (j=0; j<RACE_CHECK_DB_SIZE; j++) {

	  flags_j = race_check_db[j].flags;
	  
	  /* flags_older is set to the older of the data base entry flags word. Therefore,
	   *   If j<i  then flags_older = race_check_db[j].flags. 
	   *   If j==i then flags_older = 0. 
	   *   If j>i  then flags_older is updated as a result of further processing DMA requests.
	   */
	  if (j <= i) flags_older = flags_j;

	  if (flags_j & FLAG_INFLIGHT) {

	    /* Update flags_older to reflect ordering requests.
	     */
	    if (j > i) {
	      /* For DMAs following the one updated:
	       * 1) Flag the updated DMA as non-tag-specific barriered if a BARRIER command is found.
	       *    This allows only later atomic operations to generate a race condition error.
	       */
	      if (race_check_db[j].cmd == MFC_BARRIER_CMD) flags_older |= FLAG_BARRIERED_NON_TAG_SPECIFIC;
	      if ((tag == race_check_db[j].tag) && ((race_check_db[j].cmd & MFC_BARRIER_ENABLE) || 
						    IS_TAG_SPECIFIC_BARRIER_CMD(race_check_db[j].cmd))) {
		flags_older |= FLAG_BARRIERED_TAG_SPECIFIC;
	      }
	    }

	    if (IS_NON_QUEUED_ATOMIC_CMD(race_check_db[j].cmd) || 
		(((flags_older & FLAG_BARRIERED_NON_TAG_SPECIFIC) == 0) &&
		 ((tag != race_check_db[j].tag) || (((flags_older & FLAG_BARRIERED) == 0) && 
						    ((race_check_db[j].cmd & (MFC_BARRIER_ENABLE | MFC_FENCE_ENABLE)) == 0))))) {
	      
	      /* Check for possible race condition between two DMAs. These occur
	       * when:
	       *   two transfers overlap AND
	       *   the transfers are GET-GET, GET-PUT, or PUT-GET (PUT-PUT is ok) AND
	       * OR
	       *   previous transfer overlaps this list element array AND
	       *   previous transfer is a GET (or GETL).
	       */
	      /* Check DMA target */
	      if (((start < (race_check_db[j].start + race_check_db[j].size)) &&
		   ((start+size) > race_check_db[j].start) &&
		   (race_check_db[i].size) &&
		   IS_EITHER_GET_CMD(cmd, race_check_db[j].cmd)) ||
		  (((void *)list < (race_check_db[j].start + race_check_db[j].size)) &&
		   (((void *)list+list_size) > race_check_db[j].start) &&
		   IS_GET_CMD(race_check_db[j].cmd))) {
#if RACE_CHECK_VERBOSE
		fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for acknowledged DMA list issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d)\n",
			race_check_db[i].filename, race_check_db[i].line, cmd, start, size, tag, 
			race_check_db[j].filename, race_check_db[j].line, race_check_db[j].cmd, race_check_db[j].start, race_check_db[j].size, race_check_db[j].tag);
#else
		fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_ERROR);
#endif
	      }
	      
	      /* Check list element */
	      if (race_check_db[j].list_size) {
		if ((start < ((void *)race_check_db[j].list + race_check_db[j].list_size)) &&
		    ((start+size) > (void *)race_check_db[j].list) && 
		    (race_check_db[j].list_size) &&
		    IS_GET_CMD(cmd)) {
#if RACE_CHECK_VERBOSE
		  fprintf(RACE_CHECK_STREAM, "RACE CHECK ERROR detected for acknowledged DMA list issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d) with a list array used in the DMA issued in %s:%d (cmd=0x%x start=%p size=0x%x tag=%d)\n",
			  race_check_db[i].filename, race_check_db[i].line, cmd, start, size, tag,
			  race_check_db[j].filename, race_check_db[j].line, race_check_db[j].cmd, race_check_db[j].start, race_check_db[j].size, race_check_db[j].tag);
#else
		  fprintf(RACE_CHECK_STREAM, race_check_format, RACE_CHECK_DMAL_LIST_ERROR);
#endif
		}
	      }
	    }
	  }
	}
      }
	
      /* Unstall this DMA, marking is as in-flight
       */
      flags &= ~FLAG_STALLED;
      flags |= FLAG_INFLIGHT;
	
      race_check_db[i].flags = flags;
      break;
    }
  }
}
